from abc import ABC, abstractmethod
from collections import deque
from sys import float_info

from goap.priority_queue import PriorityQueue

__all__ = "PathNotFoundException", "AStarAlgorithm"


class PathNotFoundException(Exception):
    pass


class AStarAlgorithm(ABC):
    @abstractmethod
    def get_neighbours(self, node):
        raise NotImplementedError

    @abstractmethod
    def get_g_score(self, current, node):
        raise NotImplementedError

    @abstractmethod
    def get_h_score(self, node):
        raise NotImplementedError

    def find_path(self, start, end):
        g_scores = {start: 0}
        f_scores = {start: self.get_h_score(start)}
        parents = {}

        candidates = PriorityQueue([start], key=lambda n: f_scores[n])
        rejects = set()

        i = 0
        while candidates:
            current = candidates.pop()
            i += 1

            if self.is_finished(current, end, parents):
                return self.reconstruct_path(current, end, parents)

            rejects.add(current)

            for neighbour in self.get_neighbours(current):
                if neighbour in rejects:
                    continue

                tentative_g_score = g_scores[current] + self.get_g_score(
                    current, neighbour
                )
                tentative_is_better = tentative_g_score < g_scores.get(
                    neighbour, float_info.max
                )

                if neighbour in candidates and not tentative_is_better:
                    continue

                parents[neighbour] = current
                g_scores[neighbour] = tentative_g_score
                f_scores[neighbour] = tentative_g_score + self.get_h_score(
                    neighbour
                )

                candidates.add(neighbour)

        raise PathNotFoundException("Couldn't find path for given nodes")

    @abstractmethod
    def is_finished(self, node, goal, parents):
        raise NotImplementedError

    def reconstruct_path(self, node, goal, parents):
        result = deque((node,))

        while True:
            try:
                node = parents[node]
            except KeyError:
                break
            result.appendleft(node)

        return result



from dataclasses import dataclass
from typing import Any

from ..action import Action

State = dict[str, Any]


@dataclass(frozen=True)
class PlanStep:
    action: Action
    services: State


class Plan:

    def __init__(self, actions):
        pass



from matplotlib import pyplot as plt
from networkx import DiGraph, draw_networkx


def repr_step(step):
    return step.action.__class__.__name__


def look_ahead(iterable):
    sequence = list(iterable)
    shifted_sequence = iter(sequence)
    next(shifted_sequence)
    return zip(sequence, shifted_sequence)


def visualise_plan(plan, filename):
    graph = DiGraph()

    for i, step in enumerate(plan):
        name = repr_step(step)
        graph.add_node(name)

    if len(plan) > 1:
        for i, (step, next_step) in enumerate(look_ahead(plan)):
            name = repr_step(step)
            next_name = repr_step(next_step)
            graph.add_edge(name, next_name)

    plt.axis("off")
    draw_networkx(graph)
    plt.savefig(filename)
    plt.close()



import logging.config

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"}
    },
    "handlers": {"default": {"level": "INFO", "class": "logging.StreamHandler"}},
    "loggers": {"": {"handlers": ["default"], "level": "ERROR", "propagate": True}},
}

logging.config.dictConfig(logging_config)



from heapq import heappush, heappop


class PriorityElement:
    def __init__(self, element, score):
        self.value = element
        self.score = score
        self.removed = False

    def __lt__(self, other):
        return self.score < other.score


def _pass_through_key(x):
    return x


class PriorityQueue:
    def __init__(self, items=None, key=None):
        self._dict = {}
        self._heap = []

        if key is None:
            key = _pass_through_key

        self._key = key

        if items:
            for item in items:
                self.add(item)

    def __bool__(self):
        return bool(self._dict)

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        return iter(self._dict)

    def add(self, item):
        if item in self._dict:
            raise ValueError(f"{item} already in queue")

        element = PriorityElement(item, self._key(item))
        self._dict[item] = element
        heappush(self._heap, element)

    def pop(self):
        while True:
            element = heappop(self._heap)

            if not element.removed:
                element.removed = True
                value = element.value
                del self._dict[value]
                return value

    def remove(self, item):
        element = self._dict.pop(item)
        element.removed = True



from enum import auto, Enum
from typing import ClassVar, Dict, Any, Tuple

State = Dict[str, Any]


class ActionStatus(Enum):
    failure = auto()
    success = auto()
    running = auto()


"""
plugin(curr)  match any value (not possible without a backwards feature
callable(curr) match returned value

class SomeAction:
    effects = {
        "provide_resource": plugin()
    }
    preconditions = {
        "log_resource": ref("provide_resource")
    }

"""


class EffectReference:
    """Dynamic precondition that references an effect."""

    def __init__(self, name):
        self.goap_effect_reference = name


def reference(name: str) -> EffectReference:
    """Convenience function to return Forwarder object."""
    return EffectReference(name)


class ActionValidator(type):
    """Metaclass to validate action classes at define time."""

    def __new__(mcs, cls_name: str, bases: Tuple[type], attrs: Dict[str, Any]):
        if bases:
            # Validate precondition plugins
            preconditions = attrs.get("preconditions", {})
            # Overwrite effect plugins to ellipsis
            effects = attrs.get("effects", {})
            attrs['service_names'] = [k for k, v in effects.items() if v is Ellipsis]
            mcs._validate_preconditions(effects, preconditions)

        return super().__new__(mcs, cls_name, bases, attrs)

    @staticmethod
    def _validate_preconditions(all_effects: State, preconditions: State):
        for name, value in preconditions.items():
            if value is Ellipsis:
                raise ValueError(
                    f"Preconditions cannot be services (...): '{name}'")

            elif (
                    hasattr(value, "goap_effect_reference")
                    and value.goap_effect_reference not in all_effects
            ):
                raise ValueError(
                    f"Invalid reference name for precondition '{name}': {value.name!r}"
                )


class Action(metaclass=ActionValidator):
    effects: ClassVar[State] = {}
    preconditions: ClassVar[State] = {}
    service_names: ClassVar[list[str]] = []

    cost = 1.0
    precedence = 0.0
    apply_effects_on_exit = True

    def check_procedural_precondition(
            self, services: State, is_planning: bool
    ) -> bool:
        return True

    def get_cost(self, services) -> float:
        return self.cost



from collections import defaultdict
from dataclasses import dataclass
from operator import attrgetter
from typing import Any, Optional

from .plan import PlanStep
from ..action import Action
from ..algo import AStarAlgorithm

# Regressive algo faster
# http://alumni.media.mit.edu/~jorkin/GOAP_draft_AIWisdom2_2003.pdf


State = dict[str, Any]

_NOT_DEFINED = object()


@dataclass(frozen=True)
class RegressiveGOAPAStarNode:
    current_state: State
    goal_state: State
    action: Optional[Action] = None

    def __hash__(self):
        return id(self)

    @property
    def services(self):
        return {k: self.current_state[k] for k in self.action.service_names}

    def compute_cost(self, other: "RegressiveGOAPAStarNode") -> float:
        # The action determines the cost, and is stored on the "destination" node
        # i.e. x -- action --> y is encoded as (x, None), (y, action)
        return other.action.get_cost(other.services)

    def apply_action(self, action: Action) -> "RegressiveGOAPAStarNode":
        # New node is a consequence of transitioning from
        # node to neighbour via action (which is the edge)
        current_state = self.current_state.copy()
        for key, value in action.effects.items():
            # After applying this action, as far as the regressive planner is concerned
            # we want to our new goal to be to satisfy the actions preconditions.
            # (see the next for block)
            if value is Ellipsis:
                # This action is "fulfilling" some of our goal state, hence
                # when using services (...), we take from the goal state!
                value = self.goal_state[key]

            # Normally we work backwards from the goal by building up a current state
            # that ultimately matches our goal state.
            # However, when an action shares a precondition and effect, we effectively
            # "update" our goal locally, so we need to apply the *precondition*
            # (our "new" goal) rather than the effect (our "old" goal)

            elif key in action.preconditions:
                value = action.preconditions[key]

            current_state[key] = value

        # Now update our goals!
        goal_state = self.goal_state.copy()
        for key, value in action.preconditions.items():
            # Allow references to effects (for pass-through services)
            if hasattr(value, "goap_effect_reference"):
                value = current_state[value.goap_effect_reference]

            goal_state[key] = value

        return self.__class__(current_state, goal_state, action)

    @property
    def is_satisfied(self):
        return not self.unsatisfied_keys

    @property
    def unsatisfied_keys(self) -> list[str]:
        return [k for k, v in self.goal_state.items() if
                self.current_state.get(k, _NOT_DEFINED) != v]


_key_action_precedence = attrgetter("action.precedence")


class RegressiveGOAPAStarSearch(AStarAlgorithm):

    def __init__(self, actions: list[Action]):
        self._actions = actions
        self._effect_to_action = self._create_effect_table(actions)

    def _create_effect_table(self, actions: list[Action]):
        effect_to_actions = defaultdict(list)

        for action in actions:
            for effect in action.effects:
                effect_to_actions[effect].append(action)

        effect_to_actions.default_factory = None
        return effect_to_actions

    def reconstruct_path(self, node, goal, parents):
        path = super().reconstruct_path(node, goal, parents)
        return reversed(path)

    def is_finished(self, node: RegressiveGOAPAStarNode, goal, parents):
        return node.is_satisfied

    def get_neighbours(self, node: RegressiveGOAPAStarNode):
        neighbours = []

        for key in node.unsatisfied_keys:
            goal_value = node.goal_state[key]

            for action in self._effect_to_action[key]:
                effect_value = action.effects[key]
                # If this effect does not satisfies the goal
                if effect_value != goal_value and effect_value is not Ellipsis:
                    continue

                # Compute neighbour after following action edge.
                neighbour = node.apply_action(action)

                # Ensure we can visit this node
                if not action.check_procedural_precondition(
                        neighbour.services, is_planning=True
                ):
                    continue

                neighbours.append(neighbour)

        neighbours.sort(key=_key_action_precedence)
        return neighbours

    def get_h_score(self, node: RegressiveGOAPAStarNode):
        return len(node.unsatisfied_keys)

    def get_g_score(self, node: RegressiveGOAPAStarNode,
                    neighbour: RegressiveGOAPAStarNode):
        return node.compute_cost(neighbour)


class RegressivePlanner:
    def __init__(self, actions):
        self._actions = actions
        self._search = RegressiveGOAPAStarSearch(actions)

    def find_plan(self, current_state: State, goal_state: State):
        start = RegressiveGOAPAStarNode(current_state, goal_state, None)
        path = self._search.find_path(start, goal_state)
        plan = []
        for node in path:
            if node.action is None:
                break
            step = PlanStep(node.action, node.services)
            plan.append(step)
        return plan



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['goap', 'goap.algo', 'goap.planner']

package_data = \
{'': ['*']}

extras_require = \
{'plot': ['matplotlib>=3.0', 'networkx>=2.3']}

setup_kwargs = {
    'name': 'goap-ai',
    'version': '1.0.0',
    'description': 'Goal oriented action programming',
    'long_description': None,
    'author': 'Angus Hollands',
    'author_email': 'goosey15@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
